#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2020-03-18 11:52
@edit time: 2020-04-07 19:41
@FilePath: /vvlab/__version__.py
"""



VERSION = (0, 1, 8)

__version__ = '.'.join(map(str, VERSION))



#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2019-12-06 23:16
@edit time: 2020-01-15 16:13
@file: /wjwgym-home/vvlab/agents/__init__.py
"""


from .DDPG_base import DDPGBase
from .DQN_base import DQNBase
from .Linear_base import LinearBase


#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2019-12-04 10:40
@edit time: 2020-01-14 17:10
@file: /exp_replay.py
"""

import torch
from torch.autograd import Variable
import numpy as np


class ExpReplay:
    """
    使用类似Q-Learning算法的DRL通用的经验回放池
    1. 经验回放池的一条记录
        (state, action, reward, done, next_state) 或简写为 (s, a, r, d, s_)
        其中done的意义在于，当状态为done的时候，Q估计=r，而不是Q估计=(r + gamma * Q'max)
        显然，一条记录的维度是 dim(s) + dim(a) + dim(r) + dim(d) + dim(s_) = 2 * n_states + n_actions + 2
    2. 经验回放池的参数
        - n_states: state的维度
        - n_actions: action的维度
        - exp_size: 经验回放池的容量上限。达到这个上限之后，后续的记录就会覆盖之前的记录。
            默认值是: 1000
        - MIN_MEN: 经验回放池输出的最小数目。记录数目超过阈值之后，获取batch才会获得输出
            默认值是: exp_size//10
        - batch_size: 一个batch的大小。
            默认值是: 32
    3. 经验回放池的功能
        - 创建经验回放池对象
            >>> self.memory = ExpReplay(n_states,  n_actions, exp_size=exp_size, exp_thres=exp_thres)
        - 添加一条记录
            >>> self.memory.add_step(s, a, r, d, s_)
            将这条记录放到经验回放池的顺序位置上，例如：
                第一条记录在位置0，第2条记录在位置1，第exp_size+1条记录在位置0
        - 获取一个batch用于训练
            >>> s, a, r, d, s_ = memery.get_batch_splited()
        - 获取一个已经被转为pytorch Variable的batch用于训练
            >>> CUDA = torch.cuda.is_available()
            >>> batch = self.memory.get_batch_splited_tensor(CUDA, batch_size)
    """

    def __init__(self, n_states, n_actions, exp_size=1000, exp_thres=None, batch_size=32):
        """初始化经验回放池"""
        # 参数复制与定义
        self.n_states, self.n_actions = n_states, n_actions
        if not exp_thres:
            exp_thres = exp_size // 10
        self.exp_size, self.exp_thres = exp_size, exp_thres
        self.batch_size = batch_size
        # 初始化经验回放池
        exp_dim = n_states * 2 + n_actions + 2
        self.expreplay_pool = np.zeros((exp_size, exp_dim))
        # 初始化记录位置
        self.exp_index = 0

    def add_step(self, *step):
        """
        为经验回放池增加一步，我们约定“一步”包括s, a, r, d, s_五个部分
        将这一步的信息整合为一条记录，放置到exp_index指定的位置。之后，exp_index以exp_size为模+1
        即后来的记录会覆盖掉最早的记录

        @param *step: 接收一条需要被添加到经验回放池的记录，要求按照s, a, r, d, s_的顺序给出
        """
        # 构建 step
        s, a, r, d, s_ = step
        step = np.hstack((s.reshape(-1), a, [r], [d], s_.reshape(-1)))
        # step 添加到经验回放池
        index = self.exp_index % self.exp_size
        self.expreplay_pool[index] = step
        # 更改exp_index位置，并确保不会int溢出
        self.exp_index += 1
        if self.exp_index > 2 * self.exp_size:
            self.exp_index -= self.exp_size

    def get_batch(self, batch_size=None):
        """
        从回放池中获取一个batch
        如果经验回放池大小未达到输出阈值则返回None

        @param batch_size: 一个batch的大小，若不指定则按经验回放池的默认值
        @return 一个batch size 的记录
        """
        batch_size = batch_size if batch_size else self.batch_size
        # 超过输出阈值的时候才返回batch
        if self.exp_index >= self.exp_thres:
            expreplay_pool_size = min(self.exp_index, self.exp_size)
            choice_indexs = np.random.choice(expreplay_pool_size, batch_size)
            return self.expreplay_pool[choice_indexs]
        else:
            return None

    def get_batch_splited(self, batch_size=None):
        """
        将batch按照s, a, r, d, s_的顺序分割好并返回

        @param batch_size: 一个batch的大小，若不指定则按经验回放池的默认值
        @return cur_states, actions, rewards, dones, nexe_states: 
            按照 (s, a, r, d, s_) 顺序分割好的一组batch 
        """
        batch = self.get_batch(batch_size)
        if batch is None:
            return batch
        else:
            cur_states = batch[:, :self.n_states]
            actions = batch[:, self.n_states: self.n_states + self.n_actions].astype(int)
            rewards = batch[:, self.n_states + self.n_actions: self.n_states + self.n_actions + 1]
            dones = batch[:, -self.n_states - 1: -self.n_states].astype(int)  # 将是否结束按int类型读取，结束则为1，否则为0
            nexe_states = batch[:, -self.n_states:]
            return cur_states, actions, rewards, dones, nexe_states

    def get_batch_splited_tensor(self, CUDA, batch_size=None, dtype=torch.FloatTensor):
        """
        将batch分割并转换为tensor之后返回

        @param CUDA: 是否使用GPU，这决定了返回变量的设备类型
        @param batch_size: 一个batch的大小，若不指定则按经验回放池的默认值
        @param dtype: 返回变量的数据类型，默认为Float
        @return cur_states, actions, rewards, dones, nexe_states: 
            按照 (s, a, r, d, s_) 顺序分割好且已经转为torch Variable的一组batch 
        """
        batch = self.get_batch_splited(batch_size)
        if batch is None:
            return batch
        else:
            return (torch.from_numpy(ndarray).type(dtype).cuda() for ndarray in batch) if CUDA else (torch.from_numpy(ndarray).type(dtype) for ndarray in batch)


class OUProcess(object):
    """Ornstein-Uhlenbeck process"""

    def __init__(self, x_size, mu=0, theta=0.15, sigma=0.3):
        self.x = np.ones(x_size) * mu
        self.x_size = x_size
        self.mu = mu
        self.theta = theta
        self.sigma = sigma

    def noise(self):
        dx = self.theta * (self.mu - self.x) + self.sigma * np.random.randn(self.x_size)
        self.x = self.x + dx
        return self.x


def soft_update(target, source, tau):
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(
            target_param.data * (1.0 - tau) + param.data * tau
        )



#!/usr/bin/env python
# coding=utf-8
"""
@create time: 2019-11-22 10:19
@author: Jiawei Wu
@edit time: 2020-01-15 16:05
@file: /linear_agent.py
"""

import numpy as np
import pandas as pd


class LinearBase:
    """线性学习（QLearning、Sarsa及变形）的基类"""

    def __init__(self, actions: list, e_greedy: float, learning_rate, reward_decay):
        """
        初始化基类
        @param actions: 动作空间（注：规定动作都是非负整数）
        @param e_greedy: 选取最大价值动作的概率epsilon-greedy初始值
        @param learning_rate: 学习率
        @param reward_decay: 未来折扣率
        """
        self.actions = actions  # a list
        self.epsilon = e_greedy
        self.q_table = pd.DataFrame(columns=self.actions, dtype=np.float64)
        self.lr = learning_rate
        self.gamma = reward_decay

    def choose_action(self, observation, epsilon=None):
        """
        获取当前状态对应的动作。按照epsilon的概率从Q表读取价值最大的，剩下情况随机选取。
        @param observation: 当前状态
        @param epsilon: epsilon值，如果没有传递这个值，则按照开始设置的默认值随机。
        """
        # 确保epsilon被设置
        if not epsilon:
            epsilon = self.epsilon
        # 检测当前状态是否存在（若不存在会被初始化）
        self.check_state_exist(observation)

        # 选取动作
        if np.random.uniform() < epsilon:
            # 获取当前state对应的QTable行
            state_actions = self.q_table.loc[observation, :]
            # 在所有value最大的index中随机选取
            action = np.random.choice(state_actions[state_actions == np.max(state_actions)].index)
        else:
            # 随机选择
            action = np.random.choice(self.actions)
        return action

    def check_state_exist(self, state):
        """
        检测状态是否存在，如果不存在则添加到QTable的行，并初始化为全0
            :param self: 
            :param state: 
        """
        if state not in self.q_table.index:
            # append new state to q table
            self.q_table = self.q_table.append(
                pd.Series(
                    [0]*len(self.actions),
                    index=self.q_table.columns,
                    name=state,
                )
            )

    def print_table(self):
        """打印QTable内容"""
        q_table = self.q_table
        print(q_table)



#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2019-12-07 19:53
@edit time: 2019-12-07 19:59
@desc: DQN 使用的网络结构，其特点是输出各个离散动作值的value，
输出前会经过一个softmax函数
"""


import torch
import torch.nn as nn
import torch.nn.functional as F
CUDA = torch.cuda.is_available()


class SimpleDQNNet(nn.Module):
    """一个只有一层隐藏层的DQN神经网络"""

    def __init__(self, n_states, n_actions, n_neurons=32):
        """
        定义隐藏层和输出层参数
        @param n_obs: number of observations
        @param n_actions: number of actions
        @param n_neurons: number of neurons for the hidden layer
        """
        super(SimpleDQNNet, self).__init__()
        self.fc1 = nn.Linear(n_states, n_neurons)
        self.fc1.weight.data.normal_(0, 0.1)
        self.out = nn.Linear(n_neurons, n_actions)
        self.out.weight.data.normal_(0, 0.1)

    def forward(self, x):
        """
        定义网络结构: 第一层网络->ReLU激活->输出层->softmax->输出
        """
        if CUDA:
            x = x.cuda()
        x = self.fc1(x)
        x = F.relu(x)
        x = self.out(x)
        action_values = F.softmax(x, dim=1)
        return action_values


#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2019-12-06 23:23
@edit time: 2020-01-15 16:16
@file: /wjwgym-home/wjwgym/net_models/__init__.py
"""


from .actor_net import SimpleActorNet
from .critic_net import SimpleCriticNet
from .dqn_net import SimpleDQNNet


#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2020-04-07 17:02
@edit time: 2020-04-07 18:10
@FilePath: /vvlab/logs/assamble.py
@desc: 
"""


import json
import logging
import logging.config
import logging.handlers

levels = {
    'notset': logging.NOTSET,
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL
}

formatters = {
    'none': None,
    "simple": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    "deprecation": "%(pathname)s:%(module)s:%(funcName)s:%(lineno)d: DeprecationWarning: %(message)s"
}


def streamHandler():
    return logging.StreamHandler()


def fileHandler(filename='default.log'):
    return logging.FileHandler(filename=filename)


htypes = {
    'stream': streamHandler,
    'file': fileHandler
}


def singleHandler(htype='stream', level='warning', formatter='none', *args, **kwargs):
    # assert params are valid
    checks = {'htype': htypes.keys(), 'level': levels.keys(), 'formatter': formatters.keys()}
    for param, valids in checks.items():
        valids = list(k for k in valids)
        valids_str = ', '.join(valids[:-1]) + ' or ' + valids[-1]
        if eval(param) not in valids:
            raise ValueError(f'bad {param} (must be {valids_str})')

    # create handler
    creater = htypes[htype]
    if htype == 'file':
        filename = 'default.log' if 'filename' not in kwargs else kwargs['filename']
        handler = creater(filename)
    else:
        handler = creater()

    # set attributes
    handler.setLevel(levels[level])
    if formatter is not None:
        handler.setFormatter(logging.Formatter(formatters[formatter]))

    return handler


def singleLogger(*args, **kwargs):
    logger = logging.getLogger()
    handler = singleHandler(*args, **kwargs)
    logger.addHandler(handler)
    return logger


if __name__ == '__main__':
    logger = singleLogger(htype='stream', formatter='deprecation')
    logger.warning(123)



#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2020-04-06 15:36
@edit time: 2020-04-06 15:37
@FilePath: /vvlab/utils/OUProcess.py
@desc: 
"""

import numpy as np

class OUProcess(object):
    """Ornstein-Uhlenbeck process"""

    def __init__(self, x_size, mu=0, theta=0.15, sigma=0.3):
        self.x = np.ones(x_size) * mu
        self.x_size = x_size
        self.mu = mu
        self.theta = theta
        self.sigma = sigma

    def __call__(self):
        return self.noise()
    
    def noise(self):
        dx = self.theta * (self.mu - self.x) + self.sigma * np.random.randn(self.x_size)
        self.x = self.x + dx
        return self.x


#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2020-04-05 19:44
@edit time: 2020-04-06 15:39
@FilePath: /vvlab/utils/__init__.py
@desc: init method of utils
"""


from .config import CUDA
from .OUProcess import OUProcess
from .ReplayBuffer import ReplayBuffer

__all__ = ['CUDA', 'OUPrecess', 'ReplayBuffer']


#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2020-04-05 19:45
@edit time: 2020-04-05 20:43
@FilePath: /vvlab/utils/config.py
@desc: config file
"""

import torch
CUDA = torch.cuda.is_available()



#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2019-12-06 23:23
@edit time: 2019-12-07 20:14
@desc: Critic网络。
其特点是只需要一个输出（q值），且不需要softmax
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
CUDA = torch.cuda.is_available()


class SimpleCriticNet(nn.Module):
    """定义Critic的网络结构"""

    def __init__(self, n_states, n_actions, n_neurons=30):
        """
        定义隐藏层和输出层参数
        @param n_obs: number of observations
        @param n_actions: number of actions
        @param n_neurons: 隐藏层神经元数目
        """
        super(SimpleCriticNet, self).__init__()
        self.fc_state = nn.Linear(n_states, n_neurons)
        self.fc_state.weight.data.normal_(0, 0.1)
        self.fc_action = nn.Linear(n_actions, n_neurons)
        self.fc_action.weight.data.normal_(0, 0.1)
        self.out = nn.Linear(n_neurons, 1)
        self.out.weight.data.normal_(0, 0.1)

    def forward(self, s, a):
        """
        定义网络结构: 
        state -> 全连接   -·-->  中间层 -> 全连接 -> ReLU -> Q值
        action -> 全连接  /相加，偏置
        """
        s, a = (s.cuda(), a.cuda()) if CUDA else (s, a)
        x_s = self.fc_state(s)
        x_a = self.fc_action(a)
        x = F.relu(x_s+x_a)
        actions_value = self.out(x)
        return actions_value



#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2019-11-25 10:49
@edit time: 2020-03-30 10:12
"""

from gym.envs.registration import register

register(
    id='FindTreasure-v0',
    entry_point='vvlab.envs.findtreasure_env:FindTreasureEnv',
)
register(
    id='Maze-v0',
    entry_point='vvlab.envs.maze_env:MazeEnv',
)


#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2019-12-04 10:36
@edit time: 2020-04-07 19:48
@FilePath: /vvlab/agents/DDPG_base.py
"""
import numpy as np
import os
import logging
import torch.nn as nn
import torch
from torch.utils.tensorboard import SummaryWriter
from ..utils import CUDA, OUProcess, ReplayBuffer
import warnings


def soft_update(target, source, tau):
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(
            target_param.data * (1.0 - tau) + param.data * tau
        )


class DDPGBase(object):
    def __init__(self, n_states, n_actions, action_bound=1, buff_size=1000, buff_thres=0, batch_size=32,
                 lr_a=0.001, lr_c=0.002, tau=0.01, gamma=0.9,
                 summary=False, *args, **kwargs):
        # 兼容参数
        # TODO 在0.3.0删除参数兼容
        if 'bound' in kwargs:
            warnings.warn("'bound' is deprecated and will remove after 0.3.0. "
                          "Use 'action_bound' instead.",
                          DeprecationWarning, stacklevel=2)
            action_bound = kwargs['bound']
            self.bound = action_bound
        if 'exp_size' in kwargs:
            warnings.warn("'exp_size' is deprecated and will remove after 0.3.0. "
                          "Use 'buff_size' instead.",
                          DeprecationWarning, stacklevel=2)
            buff_size = kwargs['exp_size']
            self.exp_size = buff_size
        if 'exp_thres' in kwargs:
            warnings.warn("'exp_thres' is deprecated and will remove after 0.3.0. "
                          "Use 'buff_thres' instead.",
                          DeprecationWarning, stacklevel=2)
            buff_thres = kwargs['exp_thres']
            self.exp_thres = buff_thres

        # 参数复制
        self.n_states, self.n_actions, self.action_bound = n_states, n_actions, action_bound
        self.buff_size, self.buff_thres, self.batch_size = buff_size, buff_thres, batch_size
        self.lr_a, self.lr_c, self.tau, self.gamma = lr_a, lr_c, tau, gamma
        self.summary = summary
        self.kwargs = kwargs

        # 初始化episode和step
        self.episode, self.step = 0, 0
        # 参数覆盖
        self._param_override()

        # 创建经验回放池
        self.memory = ReplayBuffer(self.n_states, self.n_actions, buff_size=self.buff_size, buff_thres=self.buff_thres)

        # 创建神经网络
        self._build_net()
        # 指定优化器
        self.actor_optim = torch.optim.Adam(self.actor_eval.parameters(), lr=self.lr_a)
        self.critic_optim = torch.optim.Adam(self.critic_eval.parameters(), lr=self.lr_c)
        # 约定损失函数
        self.mse_loss = nn.MSELoss()

        # 指定噪声发生器
        self._build_noise()

        # 指定summary writer
        self._build_summary_writer()

        # 开启cuda
        if CUDA:
            self.cuda()

    def _param_override(self):
        """覆盖参数
        提供子类简单覆写基类参数的方法
        例如：修改summary是否开启
        应当谨慎使用这个方法
        """
        pass

    def _build_net(self):
        raise TypeError("网络构建函数未被实现")

    def _build_noise(self, *args):
        raise TypeError("噪声发生器构建函数未被实现")

    def _build_summary_writer(self):
        """构建summary_writer
        如果指定了不需要summary_writer，会将其置为None
        如果指定了保存路径就使用保存路径，否则使用默认路径
        """
        if self.summary:
            if 'summary_path' in self.kwargs:
                self.summary_writer = SummaryWriter(log_dir=kwargs['summary_path'])
                self._build_summary_writer(kwargs['summary_path'])
            else:
                self.summary_writer = SummaryWriter()
        else:
            self.summary_writer = None

    def get_summary_writer(self):
        return self.summary_writer

    def _get_action(self, s):
        """给定当前状态，获取选择的动作"""
        s = torch.unsqueeze(torch.FloatTensor(s), 0)
        action = self.actor_eval.forward(s).detach().cpu().numpy()
        return action

    def get_action(self, s):
        return self._get_action(s)

    def _save(self, save_path, append_dict={}):
        """保存当前模型的网络参数
        @param save_path: 模型的保存位置
        @param append_dict: 除了网络模型之外需要保存的内容
        """
        states = {
            'actor_eval_net': self.actor_eval.state_dict(),
            'actor_target_net': self.actor_target.state_dict(),
            'critic_eval_net': self.critic_eval.state_dict(),
            'critic_target_net': self.critic_target.state_dict(),
        }
        states.update(append_dict)
        torch.save(states, save_path)

    def save(self, episode=None, save_path='./cur_model.pth'):
        """保存的默认实现
        @param episode: 当前的episode
        @param save_path: 模型的保存位置，默认是'./cur_model.pth'
        """
        append_dict = {
            'episode': self.episode if episode is None else episode,
            'step': self.step
        }
        self._save(save_path, append_dict)

    def _load(self, save_path):
        """加载模型参数
        @param save_path: 模型的保存位置
        @return: 加载得到的模型字典
        """
        if CUDA:
            states = torch.load(save_path, map_location=torch.device('cuda'))
        else:
            states = torch.load(save_path, map_location=torch.device('cpu'))

        # 从模型中加载网络参数
        self.actor_eval.load_state_dict(states['actor_eval_net'])
        self.actor_target.load_state_dict(states['actor_target_net'])
        self.critic_eval.load_state_dict(states['critic_eval_net'])
        self.critic_target.load_state_dict(states['critic_target_net'])

        # 从模型中加载episode和step信息
        self.episode, self.step = states['episode'], states['step']
        # 返回states
        return states

    def load(self, save_path='./cur_model.pth'):
        """加载模型的默认实现
        @param save_path: 模型的保存位置, 默认是 './cur_model.pth'
        @return: 被记录的episode值
        """
        print('\033[1;31;40m{}\033[0m'.format('加载模型参数...'))
        if not os.path.exists(save_path):
            print('\033[1;31;40m{}\033[0m'.format('没找到保存文件'))
            return -1
        else:
            states = self._load(save_path)
            return states['episode']

    def _learn(self):
        """训练网络"""
        # 将eval网络参数赋给target网络
        soft_update(self.actor_target, self.actor_eval, self.tau)
        soft_update(self.critic_target, self.critic_eval, self.tau)

        # 获取batch并拆解
        batch = self.memory.get_batch_splited_tensor(CUDA, self.batch_size)
        if batch is None:
            return None, None
        else:
            self.start_train = True
        batch_cur_states, batch_actions, batch_rewards, batch_dones, batch_next_states = batch

        # 计算target_q，指导cirtic更新
        # 通过a_target和next_state计算target网络会选择的下一动作 next_action；通过target_q和next_states、刚刚计算的next_actions计算下一状态的q_values
        target_q_next = self.critic_target(batch_next_states, self.actor_target(batch_next_states))
        target_q = batch_rewards + self.gamma * (1 - batch_dones) * target_q_next   # 如果done，则不考虑未来
        # 指导critic更新
        q_value = self.critic_eval(batch_cur_states, batch_actions)
        td_error = self.mse_loss(target_q, q_value)
        self.critic_optim.zero_grad()
        td_error.backward()
        self.critic_optim.step()

        # 指导actor更新
        policy_loss = self.critic_eval(batch_cur_states, self.actor_eval(batch_cur_states))  # 用更新的eval网络评估这个动作
        # 如果 a是一个正确的行为的话，那么它的policy_loss应该更贴近0
        loss_a = -torch.mean(policy_loss)
        self.actor_optim.zero_grad()
        loss_a.backward()
        self.actor_optim.step()
        return td_error.detach().cpu().numpy(), loss_a.detach().cpu().numpy()

    def learn(self):
        c_loss, a_loss = self._learn()
        if all((c_loss is not None, a_loss is not None)):
            self.step += 1
            if self.summary_writer:
                self.summary_writer.add_scalar('c_loss', c_loss, self.step)
                self.summary_writer.add_scalar('a_loss', a_loss, self.step)

    def _add_step(self, s, a, r, d, s_):
        """向经验回放池添加一条记录"""
        self.memory.add_step(s, a, r, d, s_)

    def add_step(self, s, a, r, d, s_):
        """添加记录的默认实现
        除了添加记录之外不做任何操作
        """
        self._add_step(s, a, r, d, s_)

    def cuda(self):
        self.actor_eval.cuda()
        self.actor_target.cuda()
        self.critic_eval.cuda()
        self.critic_target.cuda()



#!/usr/bin/env python
# coding=utf-8
"""
@create time: 2019-11-22 10:21
@author: Jiawei Wu
@edit time: 2019-11-25 11:28
@file: /find_treasure.py
"""
import numpy as np
from gym.spaces import Discrete, Box
import gym

class FindTreasureEnv(gym.Env):
    def __init__(self, length=7, treasures=[6]):
        self.player_pos = 0
        self.treasures = treasures
        obs_states = [-1, 0, 1]
        self.player, self.road, self.treasure = obs_states

        self.init_obs = np.array([self.road] * length)
        for t in treasures:
            self.init_obs[t] = self.treasure
        self.observation_space = Box(shape=(length, ), low=min(obs_states), high=max(obs_states))
        self.action_space = Discrete(2)

        head_left, head_right = 0, 1
        ac_states = [head_left, head_right]

        self.step_count = 0
    
    def reset(self):
        self.player_pos = 0
        obs = self.init_obs.copy()
        obs[self.player_pos] = self.player
        self.step_count = 0

        return obs

    def get_obs(self):
        obs = self.init_obs.copy()
        obs[self.player_pos] = self.player
        return obs
        
    def get_reward(self):
        move, win = 0, 1
        if any((self.player_pos == t_pos for t_pos in self.treasures)):
            return win
        else:
            return move

    def get_done(self):
        if any((self.player_pos == t_pos for t_pos in self.treasures)):
            return True
        else:
            return False
            
    def get_info(self):
        self.step_count += 1
        return self.step_count
        
    def step(self, action):
        if action == 0:
            delta = -1
        else:
            delta = 1
        next_pos = self.player_pos + delta
        if 0 <= next_pos < self.observation_space.shape[0]:
            self.player_pos = next_pos
 
        obs = self.get_obs()
        reward = self.get_reward()
        done = self.get_done()
        info = self.get_info()
        return obs, reward, done, info

    def render(self):
        ascii_map = {
            -1: '*',
            0: '-',
            1: 'o'
        }
        print(f'{"".join([ascii_map[d] for d in self.get_obs()])}\r', end='')




    


#!/usr/bin/env python
# coding=utf-8
"""
@create time: 2019-11-22 15:01
@author: Jiawei Wu
@edit time: 2019-11-26 16:16
@file: /maze.py
"""

import numpy as np
from gym.spaces import Discrete, Box
import gym


class MazeEnv(gym.Env):
    def __init__(self, height=4, width=4, hells=[[1, 2]], ovals=[[2, 1]]):
        """
        初始化迷宫环境，长宽都是5
        """
        self.road, self.player, self.hell, self.oval = 0, 1, 2, 3   # 定义状态
        self.action_space = Discrete(4)
        self.observation_space = Box(low=-1, high=1, shape=(height, width))
        init_space = np.zeros((height, width))
        self.player_pos = np.array([0, 0])
        self.hells, self.ovals = hells, ovals
        for h in hells:
            x, y = h
            init_space[x][y] = self.hell
        for o in ovals:
            x, y = o
            init_space[x][y] = self.oval
        self.init_space = init_space    # 初始化状态
        self.step_count = 0     # 初始化步数记录

    def reset(self):
        """重置迷宫环境"""
        self.player_pos = np.array([0, 0])
        self.step_count = 0
        return self.get_obs()

    def get_obs(self):
        """获取当前状态"""
        obs = self.init_space.copy()
        x, y = self.player_pos
        obs[x][y] = self.player
        return obs

    def get_reward(self):
        """获取奖励"""
        death, move, win = -1, 0, 1  # 定义不同状态的奖励
        if any(all(self.player_pos == h) for h in self.hells):
            return death
        if any(all(self.player_pos == o) for o in self.ovals):
            return win
        return move

    def get_done(self):
        """获取是否结束"""
        return any(all(self.player_pos == h) for h in self.hells) or any(all(self.player_pos == o) for o in self.ovals)

    def get_info(self):
        """获取额外信息"""
        return self.step_count

    def step(self, action):
        """
        执行一步动作，在迷宫环境中是向给定的方向走一步
        我们规定[up down left right] 分别为 [0 1 2 3]
        @param: action 传入的动作
        """
        action = np.clip(int(action), 0, 3)  # 限制action是整数且范围不能超过range(4)
        actions = {
            0: np.array([-1, 0]), 1: np.array([1, 0]), 2: np.array([0, -1]), 3: np.array([0, 1])
        }   # 定义action对应的x和y值变化
        next_pos = self.player_pos + actions[action]
        h_limit, w_limit = self.observation_space.shape
        h_pos, w_pos = next_pos
        if all((0 <= h_pos < h_limit, 0 <= w_pos < w_limit)):
            # 只有x和y都在范围内才算这一步有效
            self.player_pos = next_pos
        self.step_count += 1
        return self.get_obs(), self.get_reward(), self.get_done(), self.get_info()

    def render(self):
        pass



#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2019-11-17 11:23
@edit time: 2020-01-14 17:25
@file: /dqn.py
@desc: 创建DQN对象
"""
import torch
import torch.nn as nn
import numpy as np
import pandas as pd
import gym
from .Utils import ExpReplay
CUDA = torch.cuda.is_available()


class DQNBase(object):
    def __init__(self, n_states, n_actions, learning_rate=0.01, discount_rate=0.9):
        """
        初始化DQN的两个网络和经验回放池
        @param n_obs: number of observations
        @param n_actions: number of actions
        """
        # DQN 的超参
        self.gamma = discount_rate  # 未来折扣率

        self.epsilon = 0.9
        self.epsilon_min = 0.01
        self.epsilon_decay = 0.9
        self.eval_every = 10
        # 网络创建
        self.n_states, self.n_actions = n_states, n_actions
        self._build_net()

        self.replay_buff = ExpReplay(n_states, 1, exp_size=200)
        # 定义优化器和损失函数
        self.optimizer = torch.optim.Adam(self.eval_net.parameters(), lr=learning_rate)
        self.loss_func = nn.MSELoss()
        # 记录步数用于同步参数
        self.eval_step = 0
        if CUDA:
            self.cuda()

    def _build_net(self):
        raise TypeError("Network build no implementation")

    def get_action(self, state):
        # epsilon update
        self.epsilon = self.epsilon * self.epsilon_decay if self.epsilon > self.epsilon_min else self.epsilon
        # 将行向量转为列向量（1 x n_states -> n_states x 1 x 1)
        if np.random.rand() < self.epsilon:
            # 概率随机
            return np.random.randint(self.n_actions)
        else:
            # greedy
            state = torch.unsqueeze(torch.FloatTensor(state), 0)
            action_values = self.eval_net.forward(state).cpu()
            return action_values.data.numpy().argmax()

    def get_raw_out(self, state):
        state = torch.unsqueeze(torch.FloatTensor(state), 0)
        action_values = self.eval_net.forward(state)
        return action_values

    def add_step(self, cur_state, action, reward, done, next_state):
        self.replay_buff.add_step(cur_state, action, reward, done, next_state)

    def learn(self):
        batch = self.replay_buff.get_batch_splited_tensor(CUDA)
        if batch is None:
            return
        # 参数复制
        if self.eval_step % self.eval_every == 0:
            self.target_net.load_state_dict(self.eval_net.state_dict())
        # 更新训练步数
        self.eval_step += 1
        # 拆分batch
        batch_cur_states, batch_actions, batch_rewards, batch_dones, batch_next_states = batch
        # 计算误差
        q_eval = self.eval_net(batch_cur_states)
        q_eval = q_eval.gather(1, batch_actions.long())  # shape (batch, 1)
        q_next = self.target_net(batch_next_states).detach()     # detach from graph, don't backpropagate
        # 如果done，则不考虑未来
        q_target = batch_rewards + self.gamma * (1 - batch_dones) * q_next.max(1)[0].view(len(batch_next_states), 1)   # shape (batch, 1)
        loss = self.loss_func(q_eval, q_target)
        # 网络更新
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        return loss.detach().cpu().numpy()

    def cuda(self):
        self.eval_net.cuda()
        self.target_net.cuda()



#!/usr/bin/env python
# coding=utf-8
"""
@author: Jiawei Wu
@create time: 2019-12-06 23:23
@edit time: 2019-12-12 11:11
@desc: DDPG中Actor使用的网络
特点是有一个bound的特殊参数。因为Actor往往只输出一个动作，但是有上下限.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
CUDA = torch.cuda.is_available()


class SimpleActorNet(nn.Module):
    """定义Actor的网络结构"""

    def __init__(self, n_states, n_actions, n_neurons=30, a_bound=1):
        """
        定义隐藏层和输出层参数
        @param n_obs: number of observations
        @param n_actions: number of actions
        @param n_neurons: 隐藏层神经元数目
        @param a_bound: action的倍率
        """
        super(SimpleActorNet, self).__init__()
        self.bound = a_bound
        self.fc1 = nn.Linear(n_states, n_neurons)
        self.fc1.weight.data.normal_(0, 0.1)
        self.out = nn.Linear(n_neurons, n_actions)
        self.out.weight.data.normal_(0, 0.1)
        if CUDA:
            self.bound = torch.FloatTensor([self.bound]).cuda()
        else:
            self.bound = torch.FloatTensor([self.bound])

    def forward(self, x):
        """
        定义网络结构: 第一层网络->ReLU激活->输出层->tanh激活->softmax->输出
        """
        x = x.cuda() if CUDA else x
        x = self.fc1(x)
        x = F.relu(x)
        x = self.out(x)
        action_value = F.tanh(x)
        action_value = action_value * self.bound
        return action_value

